import time
from tqdm import tqdm
from types import FunctionType
from crawlfish.http import get_url 
from .url_to_folder_name import url_to_folder_name
from .website_report import WebsiteReport
from .explore_webpage import explore_webpage
from .ensure_html_url import ensure_html_url

def explore_website(  
	url:str   ,
	get_function:FunctionType = get_url  ,
	crawl_limit:int = 0
	   ):
	''' Explores all links of a website stemming from the url provided. It will try to discover new parts of the website
	Parameters
	----------
	url:str
		The url to the website to be explored
	get_function:FunctionType
		A function for fetching the webpage ,
		it should take a url as its sole argument and return a html string or BeautifulSoup instance
	crawl_limit:int
		The maximum number of pages to be crawled

	Returns
	-------
	WebsiteReport
		A WebsiteReport instance containing all the scraped data 

	'''
	crawl_limit_check = True if crawl_limit != 0 else False 
	page_reports = {}
	url_pool = [url, ]
	fetched_urls = [  ] 
	discarded_urls = []
	current_url = url 
	count = 0 
	while url_pool != [] :

		print("=================== CRAWLING {} URL | CRAWL COUNUT =  {} | POOL SIZE = {} ==================".format(current_url, count , len(url_pool)))
		time.sleep(1)
		if crawl_limit_check:
			if count >= crawl_limit:
				print("Crawl limit reached ! ")
				time.sleep(1)
				break 
		if current_url in discarded_urls:
			if current_url in url_pool:
				url_pool.pop(url_pool.index(current_url))
			count += 1
			continue
		if not ensure_html_url(current_url ):
			if current_url in url_pool:
				url_pool.pop(url_pool.index(current_url))
			discarded_urls.append(current_url)
			count += 1
			continue
		if current_url in fetched_urls:
			count += 1
			if current_url in url_pool:
				url_pool.pop(url_pool.index(current_url))
				continue
			continue
		page_report = explore_webpage(current_url , get_function)
		folder_name = url_to_folder_name(current_url)
		page_reports[folder_name] = page_report
		fetched_urls.append(current_url)
		url_pool.pop(url_pool.index(current_url))
		urls = page_report.all_site_urls
		for u in urls :
			if u in fetched_urls:
				continue
			url_pool.append(u)
		url_pool = list(set(url_pool))
		for u in url_pool:
			if u in fetched_urls:
				url_pool.pop(url_pool.index(u))
		if len(url_pool) > 0:
			current_url = url_pool[0]
		count += 1
	report = WebsiteReport(page_reports)
	for pr in page_reports.values():
		print(pr)
	print("Discarded {} urls as they were not HTML page urls ".format(len(discarded_urls)))
	return report



from crawlfish import ElementFinder


# Find all <h2></h2> elements from a html string 
html = "<h2>Hello</h2> <h2>there </h2> <h2>you</h2>"
finder = ElementFinder(html)
elements = finder.find_elements(tag_name = "h2") 

# Find just 2 <h2> elements from a html string
html = "<h2>Hello</h2> <h2>there </h2> <h2>you</h2>"
finder = ElementFinder(html)
elements = finder.find_elements(tag_name = "h2" , num = 2) 


# Find  elements by the tag name , atribute name and attribute value  
html = """
<!DOCTYPE html>
<html>
<head>
	<title> Let's find elements </title>
</head>
<body>

<div></div>
<div></div>
<div></div>
<div id="name" >Sherlock Holmes</div>
<div id="name" >Mansa Musa</div>



</body>
</html>

"""
finder = ElementFinder(html)
elements = finder.find_elements( tag_name = "div",  attribute = "id" , attribute_value = "name"  )



# Search elements by text match
# normal text match example
html = '<div id="person"> @ ceo Zuck </div> <div id="person">@warlord Khan</div> <div id="person">@ ceo John</div>'
finder = ElementFinder(html)
elements = finder.find_elements( text_match = "ceo" )
elements

# with regex example
html =  '<div>This is not an email</div> <div>randomemail@yahoo.com</div> <div>anotheremail@gmail.com</div>'
finder = ElementFinder(html)
email_regex = r"[\w\.-]+@[\w\.-]+(\.[\w]+)+" 
elements = finder.find_elements(text_match = email_regex)
elements



# Search for elements by multiple values 
html = """
<!DOCTYPE html>
<html>
	<head>
		<title> Let's find elements </title>
	</head>
	<body>

	<div id="person" class="person-container" > 
		 Someone had an email and it was someemail@gmail.com , yeah , his name was Ail Someem
	</div>

	<div id="" class="" ></div>
	<div id="" class="" ></div>
	<div id="" class="" ></div>

	<div id="" class="email-div" ></div>

	<p id="email-p" >validemail@gmail.com</p>
	<p id="email-p" ></p>
	<p id="" class="email-p"></p>
	<p id="" class="email-p"></p>
	<div id="" class="email-div" >Anothervalidemail@gmail.com</div>
	<div id="" class="email-div" ></div>

	<p id="" class="email-p"></p>

	<a id="" class="" ></a>
	<a id="" class="" ></a>
	<a id="" class="" ></a>
	<a id="" class="" ></a>
	<a id="" class="" ></a>



	</body>
</html>

"""
# Lets collect elements with valid email addresses as an example
# You can see the data is jumbled up but there are some constant information we can utilize to get everything in one line
finder = ElementFinder(html)
email_regex = r"[\w\.-]+@[\w\.-]+(\.[\w]+)+" 
elements = finder.find_elements(  attributes = [ "id" , "class"] , attribute_values = ['email' , ] ,
		 tag_names = [ 'div' , 'p'] ,text_matches = [email_regex, ]    )
elements



class InvalidSaveOptionError(Exception):
	''' To be raised when an object is not a SaveOption object '''
	def __init__(self , message):
		self.message = "Crawlfish could not load this SaveOption class -> " +  message
		super().__init__(self.message)


 
class UnsupportedSaveOptionInstanceError(Exception):
	''' To be raised when the  SaveOption instance provided is not of a supported type
	  '''
	def __init__(self , save_option_instance , supported_save_option_instances):
		self.message = """
		The SaveOption instance provided [{}] is an unsupported class of SaveOption . 
		 Supported classes are : 
		 -----------------------
		 {}
		 ------------------------
		 """.format(save_option_instance , supported_save_option_instances)
		super().__init__(self.message)


class InvalidSaveOptionError(Exception):
	''' To be raised when the  static type provided isn't supported
	  '''
	def __init__(self ,type_provided):
		self.message = """Invalid static type {} . Only Javascript and CSS are supported """.format(type_provided)
		super().__init__(self.message)





from .save_options import  CSVSaveOption , ExcelSaveOption , JSONSaveOption,JSONFileSaveOption

from .download_static_code_file import download_static_code_file , download_static_code_files
from .download_image import download_image ,download_images
from .download_video import download_video ,download_videos
from .explore_webpage import explore_webpage
from .explore_website import explore_website
from .extract_file_info_from_url import extract_file_info_from_url
from .ensure_html_url import ensure_html_url
from .crawler import UrlsCrawler
from .save import ListDataSaver
from .website_report import WebsiteReport
from .webpage_report import WebpageReport
from .save_options import ExcelSaveOption,CSVSaveOption, JSONFileSaveOption , JSONSaveOption  


__all__ = [
	
	'download_static_code_file',
	'download_static_code_files',
	'download_video',
	'download_videos',
	'download_image',
	'download_images',
	'explore_webpage',
	'explore_website',
	'extract_file_info_from_url',
	'ensure_html_url',
	'UrlsCrawler',
	'ListDataSaver',
	'WebsiteReport',
	'WebpageReport',
	'ExcelSaveOption',
	'JSONSaveOption',
	'JSONFileSaveOption',
	'CSVSaveOption',
]





from urllib.parse import urlparse  # Python 3.x
from .get_base_url import get_base_url




def get_url_parts(url:str):
	'''This function splits a url and returns  possible routes of the url

	Example :
		Https://site.com/a/b/c/d

		to return 	[ https://site.com/a/ , https://site.com/a/b,https://site.com/a/b/c ]

	Parameters
	----------
	url:str
		The url to extract the possible url parts from

	Returns
	-------
	parts:list
		All possible url parts found
	'''
	path = urlparse(url).path[1:]
	base_url = get_base_url(url)
	parts = path.split('/')
	parts = [base_url+'/' + '/'.join(parts[:index+1]) for index in range(len(parts))]
	if url in parts:
		parts.pop(parts.index(url))
	return parts 




from bs4 import BeautifulSoup

from crawlfish.html import ensure_soup , ElementFinder

def extract_html_meta(html_or_soup:tuple[BeautifulSoup,str]):
	''' This function extracts meta data information from a html page or bs4.BeautifulSoup  object and returns it in a dictionary

	Parameters
	----------
	html_or_soup:[str , BeautifulSoup]
		The html string or soup instance from where to extract the metadata from 
	
	Returns
	-------
	data:dict
		The metadata extracted in the format ->   { key:value..... }
	'''
	soup = ensure_soup(html_or_soup)
	finder = ElementFinder(soup)
	tags = soup.find_all("meta")
	data = {}
	attribute_combinations =[
			['name' ,'content'], 
			]
	for tag in tags:
		for combination in attribute_combinations:
			key_attr = combination[0]
			value_attr = combination[1]
			if tag.has_attr(key_attr) and tag.has_attr(value_attr):
				data[tag[key_attr]]=tag[value_attr]
	return data 



from crawlfish import explore_webpage , explore_website 


# Lets expore a web page 
page_report = explore_webpage("https://stackoverflow.com")
# returns a crawlfish.WebpageReport object 




# Save the scraped data 
page_report.save( output_folder = "mydata" )
# a folder named my data will be created and youll find all the data there 

# For various reasons  ,you may want to prevent downloading various files like images or css 
# you can set that like so :
page_report.save("mydata" , download_js = True , download_css = False  , download_images = False )
# in the output  folder , there will be no images or css , only Javascript

# access the data to do something 
page_report.js_urls # urls to javascript files
page_report.css_tags # bs4.Tag instances 

help(page_report) # for more info 


# Lets explore a website 
# The 'crawl_limit' kwarg sets the maximum number of pages to crawl , default is 0 which equals to no limit
website_report = explore_website("https://stackoverflow.com"  , crawl_limit = 3 )
# website_report is a crawlfish.WesiteReport instance
# save the report 
website_report.save(output_folder = "WebsiteReport")
# The data will be found in the output folder provided

# For more help :
help(website_report)
help(explore_website)




from .test_element_finder import *
from .test_get_element_text import *
from .test_get_first_text_value import *
from .test_get_first_attr_value import *
from .test_http_get import *
from .test_get_base_url import *
from .test_ensure_html_url import *
from .test_get_url_parts import *
from .test_extract_file_info_from_url import *
from .test_webpage_report import *
from .test_website_report import *
from .test_scraped_data_saver import *
from .test_explore_website import *
from .test_explore_webpage import *
from .test_urls_crawler import *
from .test_download_static_code_file import *
from .test_download_image import *
from .test_download_video import *

from .test_ensure_url_format import *
from .test_extract_html_meta import *

from .test_url_to_folder_name import *

from .test_check_file import *
from .test_get_page_with_driver import *


import requests
import os
from types import FunctionType
from urllib.request import urlopen
from crawlfish.crawlfish_utils import check_folder ,check_file
from crawlfish.http import get_url
from .extract_file_info_from_url import extract_file_info_from_url

def download_image(
	url:str ,
	get_function:FunctionType = get_url ,
	file:str = '' ,
	output_folder:str = os.getcwd() , 
	file_name:str = '' ,
	save_to_file:bool = True
	):

	'''Downloads an image file from a url , returns the image data and saves the image to a file if save_to_file=True

	Parameters
	----------
	url:str
		The URL of the image to want to download
	get_function:FunctionType
		A function that takes the url as a parameter and returns a request.Response object 
	file:str
		The name of the file to write the downloaded data to 
	output_folder:str
		The folder/directory where to store the downloaded file 
	save_to_file:bool
		Whether or not to save the data to a file 
	file_name:
		The name of the file withholding the extension name where the downloaded data will be written .
		The extension will be added automatically if this is provided 


	Returns
	-------
	tuple 
		(file_name,image_data)


	'''
	response = get_function(url)
	content =response.content
	url_file_name , extension = extract_file_info_from_url(url)
	if output_folder:
		output_folder = os.path.abspath(output_folder)
		check_folder(output_folder)
	if output_folder and not file and not file_name:
		file =os.path.join(output_folder , url_file_name)
	if file_name:
		file_name  = file_name + extension
		file = os.path.join(output_folder , file_name)
	if not file:
		file  = os.path.join(output_folder , url_file_name)
	if not file_name:
		file_name = os.path.split(file)[-1]
	if save_to_file:
		print("Saving image data to file ------- ", file ,' ', end = '')
		if not check_file(file  ):
			print("Could not save to file :( " , file)
			return file_name , ''
		with open(file , "wb" ) as f:
			f.write(content)
		print("DONE")	
	return file_name , content

def download_images(urls_list:tuple[list,tuple,iter,set], get_function = get_url ,
						 output_folder = os.getcwd()):
	'''Downloads images from a list of urls and returns the content and data of the images in a list 
	

	Parameters
	----------
	urls_list:list,tuple,iter,set
		An iterable containing the uruuls of the images to be downloaded
	get_function:FunctionType
		A function that takes a url as an argument and returns a request.Response object 
	output_folder:str
		The folder/directory where to store the downloaded files

	Returns
	-------
	image_data:dict
		{ filename:bytes_data , ...... }
		A dictionary containing the image data as values and file names as keys
	'''

	image_data ={}
	for url in urls_list:
		file_name , data = download_image(url , get_function = get_function , output_folder = output_folder)
		image_data[file_name] = data 
	return image_data




from crawlfish.crawler.exceptions import InvalidSaveOptionError  
from crawlfish.data_saving_formats import save_formats




class SaveOption:
	"""
	A base class for specifying how scraped data is to be saved
	"""
	save_format = ""
	def __init__(self):
		pass 


class CSVSaveOption(SaveOption):
	def __init__(self , file:str , overwrite:bool = True , delimiter:str=','):
		SaveOption.__init__(self)
		self.file = file 
		self.overwrite = overwrite
		self.save_format = save_formats.CSV_FILE
		self.delimiter = delimiter


 


class ExcelSaveOption(SaveOption):
	def __init__(self , file:str , sheet_name:str , overwrite:bool = True):
		SaveOption.__init__(self)
		self.file = file 
		self.sheet_name = sheet_name 
		self.overwrite = overwrite
		self.save_format = save_formats.EXCEL_FILE

class JSONSaveOption(SaveOption):
	def __init__(self , key_index:int = 0 ):
		self.save_format = save_formats.JSON 
		self.key_index = key_index 

class JSONFileSaveOption(SaveOption):
	def __init__(self ,file ,  key_index:int = 0 ,
				overwrite:bool = True,
				 ensure_ascii:bool = False  ,
				 indent:int = 4 , 
				 encoding:str = 'utf-8' ):
		self.save_format = save_formats.JSON_FILE
		self.key_index = key_index 
		self.ensure_ascii = ensure_ascii
		self.encoding = encoding
		self.indent = indent
		self.file = file 
		self.overwrite = overwrite
		


 



from .get_element_text import get_element_text 
from .get_elements_text import get_elements_text 
from .element_finder import ElementFinder
from .ensure_soup import ensure_soup
from .get_first_attr_value import get_first_attr_value
from .get_first_text_value import get_first_text_value


__all__ = [
	
	"ensure_soup", 
	"ElementFinder" , 
	"get_elements_text",
	"get_element_text",
	"get_first_text_value", 
	"get_first_attr_value",

	]


import os 


def check_folder(folder:str):
	''' CHecks if a folder exists and tries to create if it if not .Returns an error if the foldr couldnt be created


	Parameters
	----------
	folder:str
		The name of the folder to check 



	Returns
	-------
	bool
		Whether or not  the folder finally  exists or not 
		'''

 
	if not os.path.isdir(folder)  :
		try:
			os.makedirs(folder)
			return True
		except:
			raise 
		return False
	return True


import os 


def extract_file_info_from_url(url:str):
	'''This funtion extracts the name and extension name of a file from a url 

	'https:site.com/file.mp4?key=djhfjddata=45'

		will return :

	 'file.mp4' , '.mp4' 

	 Parameters
	 ----------
	 url:str
	 	The url to extract from"""
	Returns
	-------
	tuple
		The data extracted in this format: 
			(file_name , extension)	
	'''
	split = url.split("/")
	file_name =split[-1]
	file_name = file_name.split("?")[0]
	extension = os.path.splitext(file_name)[-1]
	return file_name , extension


import os
from collections.abc import Iterable
from datetime import datetime
from types import FunctionType
from tabulate import tabulate
from tqdm import tqdm
from .url_to_folder_name import url_to_folder_name
from crawlfish import get_url


SUMMARY_REPORT = """

Crawlfish has successfully scraped all {} links 

I'm working on this summary (I'm lazy) so more is to come in the next versions ....................

"""

class WebsiteReport:
	''' This class is to be used as a contaoiner for all the data collected during website crawling 

	Parameters
	----------
	page_reports:dict
		A dictionary containing page reports. Format -> { folder_name:WebpageReport }
	get_function:FunctionType
		A function that takes a url argument and returns a requests.Response object

	'''

	def __init__(self , page_reports:dict , get_function:FunctionType = get_url):
		self.page_reports = page_reports
		self.js_files_num =0 
		self.css_files_num = 0 
		self.images_num = 0 
		self.other_site_urls = []
		self.elements_processed = 0
		self.get_function = get_function
		self.process_stats()

	def __repr__(self):
		text = "{}".format(tabulate( [
				["Pages crawled" ,len(self.page_reports.keys())],
				["Other site urls found", len(self.other_site_urls)],
				["CSS files found" , self.css_files_num] , 
				["Javascript files found" , self.js_files_num],
				["Image files found", self.images_num],
				["HTML tags processed" ,self.elements_processed],
				]
				))
		return text 
	def __str__(self):
		return self.__repr__()


	def process_stats(self):
		''' Prepare and compute necessary stats and class attributes '''
		for pr in self.page_reports.values():
			self.images_num += len(pr.image_urls)
			self.js_files_num += len(pr.js_urls)
			self.css_files_num += len(pr.css_urls)
			self.other_site_urls += pr.other_site_urls
			self.elements_processed += pr.elements_processed

	def  save(self  , output_folder = os.getcwd()  , download_images:bool = True ,
				 download_js:bool =True , download_css:bool= True , redownload_static:bool=False , 
				 exclude_static:Iterable = [] , name_space:bool = True ):
		''' Save the website information obtained by saving the web report classes and saving general website information 

		Parameters
		----------
		output_folder:str
			The folder where to save the data 
		download_images:bool
			Whether or not to download images
		download_js:bool
			Whether or not to download external Javascript static files  
		download_css:bool
			Whether or not to download external CSS static files  
		name_space:bool
			Whether or not to uniquely namespace the output folder using a time tag eg. OutputFolder_3rd_March/
		redownload_static:bool
			Whether or not to redownload static files that have already been downloaded instead of caching the data
		exclude_static:Iterable 
			A list of static file types to NOT download 
		name_space:bool 
			Whether or not to name space the page report save folders 


		Returns
		-------
		None
		'''
		now = datetime.now()
		static_urls_processed = [] 
		static_data_downloaded = {}
		string_time = now.strftime("%H_%M_%A_%d_%Y")
		output_folder  = os.path.join(output_folder , "Website_report_"+  string_time)
		site_report_file = os.path.join(output_folder , "Site Report.txt")
		other_site_urls_file =os.path.join(output_folder , "Other site urls.txt")
		summary_report_file =os.path.join(output_folder , "Summary.txt")

		for folder_name , page_report in tqdm(self.page_reports.items(), "Saving page report --- "):
			page_report_folder = os.path.join(output_folder,folder_name)
			page_report.get_function = self.get_function
			if redownload_static:
				page_report.already_downloaded_static = static_data_downloaded
			download_images  ,download_js , download_css = False if 'img' in exclude_static else True  , False if 'js' in exclude_static else True  , False if 'css' in exclude_static else True  

			page_report.save(output_folder = page_report_folder , download_images = download_images , download_js = download_js , download_css= download_css , name_space = name_space)
			static_data_downloaded.update(page_report.all_static_data)



		with open(site_report_file , "w") as f:
			f.write(self.__repr__())

	 


from crawlfish import download_images , download_static_code_files , download_video 
from crawlfish import download_image , download_static_code_file , download_videos 

# These are some pretty common functions 
# use the 'save_to_file' keyword in all of these functions to choose 
# whether to save to a file or not


# Download an image
file_name , bytes_data = download_image( 'https://www.sample-videos.com/img/Sample-jpg-image-50kb.jpg' , output_folder = "test_data/myimages/" )
# The image will be saved to the output folder
file_name
# Download images from a list of urls
images = ['https://www.sample-videos.com/img/Sample-jpg-image-50kb.jpg','https://www.sample-videos.com/img/Sample-jpg-image-100kb.jpg']
data =  download_images(images , output_folder = "test_data/my_images") # { file_name:bytes_data.... }
data.keys()

# lets not save to a file by doing this
file_name , bytes_data = download_image( 'https://www.sample-videos.com/img/Sample-jpg-image-50kb.jpg' , save_to_file = False)



# Download a video 
file_name , bytes_data = download_video('https://www.sample-videos.com/video321/mp4/720/big_buck_bunny_720p_2mb.mp4' , output_folder = "test_data/")
# Download videos
videos  = ['https://www.sample-videos.com/video321/mp4/720/big_buck_bunny_720p_2mb.mp4' , 'https://www.sample-videos.com/video321/mp4/720/big_buck_bunny_720p_5mb.mp4']
data = download_videos(videos )



# Download a text file 
url = 'https://ajax.googleapis.com/ajax/libs/jquery/3.7.1/jquery.min.js'
file_name , text = download_static_code_file(url)
file_name

# For more info use the help function to read the docstrings



import os
from types import FunctionType
from collections.abc import Iterable
from tqdm import tqdm
from crawlfish.crawlfish_utils import check_file
from .save import ListDataSaver
from .save_options import save_formats, SaveOption ,ExcelSaveOption  
from .save_options import  JSONSaveOption , JSONFileSaveOption, CSVSaveOption 

default_csv_save_option = CSVSaveOption("ScrapedData.csv")

def check_save_option(save_option , crawler_object):
	''' Checks if a saveoption instance is supported , raises an Error of not'''
	supported = crawler_object.supported_save_option_instances
	if not save_option in supported:
		raise UnsupportedSaveOptionInstanceError(save_option , supported)

class Crawler:
	'''This is the base class for crawling
	'''
	supported_save_option_instances = CSVSaveOption , JSONSaveOption , JSONFileSaveOption , ExcelSaveOption
	output_folder = os.getcwd()
	def __init__(self , save_option = default_csv_save_option):
		self.save_option = save_option

	  

class UrlsCrawler(Crawler):
	''' This Crawler is used as a wrapper for  scraping  a list of urls using a function that takes the url as a parameter and returns data from one url'''

	def __init__(self , save_option:SaveOption = default_csv_save_option ):
		Crawler.__init__(self , save_option)



	def scrape_urls_in_file(
		self ,
		file:str ,
		scrape_function:FunctionType ,
		data_headers:Iterable ,  
						 save_option:SaveOption = None ):
		'''This method scrapes urls that are found in the text based file provided

		Parameters
		----------
		file:str
			The name of the file containing the urls 
		scraper_function:FunctionType
			A function that takes a url as it's sole parameter and returns a list of items scraped from the website . 
		data_headers:Iterable
			The headers for the data to be scraped eg. [NAME , AGE , SUBJECT ....]
			The number of items should be the same length as the list returned by the scrape_function
		save_option:SaveOption
			A SaveOption instance that specifies how the data should be saved 

		Returns
		-------
		scraped_data:list
			A list of the data scraped in this format : [  data_headers , [ row1_column1 , row1_colum2 ] , [ row2_column1 , row2_column2 ].....]
		'''
		
		if not save_option:
			save_option = self.save_option
		if not check_file(file):
			raise FileNotFoundError("Could not find the file [{}]".format(file))
		urls = [] 
		with open(file , 'r') as f:
			urls = f.readlines()
		return self.scrape_urls(urls , scrape_function , 
								data_headers, save_option  )

		
	def scrape_urls(self , urls:tuple[list , tuple ,set , iter ] , scrape_function:FunctionType , data_headers:tuple[list, tuple , set, iter],  
					 
						 save_option:SaveOption = None ):
		'''Scrapes a list of urls using the scrape function provided and returns the data 
		Parameters
		----------
		urls:[list,tuple ,set , iter ]
			The iterable containing the urls you want to scrape 
		scraper_function:FunctionType
			A function that takes a url as it's sole parameter and returns a list of items scraped from the website . 
		data_headers:Iterable
			The headers for the data to be scraped eg. [NAME , AGE , SUBJECT ....]
			The number of items should be the same length as the list returned by the scrape_function
		save_option:SaveOption
			A SaveOption instance that specifies how the data should be saved 	

		Returns
		-------
		scraped_data:list
			A list of the data scraped in this format : [  data_headers , [ row1_column1 , row1_colum2 ] , [ row2_column1 , row2_column2 ].....]
		

		'''
		if not save_option:
			save_option = self.save_option
		scraped_data = []
		for url in tqdm(urls):
			data = scrape_function(url)
			scraped_data.append(data)
		# add headers to the data 
		scraped_data_to_write = [data_headers, ] + scraped_data
		data_saver = ListDataSaver()
		data_saver.save_by_option(scraped_data_to_write , save_option)
		return scraped_data

 
 
 


class save_formats:
	'''
	For representing the save format for scraped data 
	The attributes just act as quick to access identifiers for the save type when only the save type is needed


	''' 

	EXCEL_FILE = "EXCEL_SPREADSHEET_FORMAT"
	CSV_FILE = "CSV_FORMAT"
	JSON  = "jSON OBJECT"
	JSON_FILE = "JSON FILE " 
	TEXT_FILE= "TEXT BASED FILE"
	IMAGE = "PNGs, JPEGS ....."



import os
import urllib.parse 


def ensure_url_format(
	base_url:str ,
	url:str
	):
	''' Ensures that the url is full and starts with the base_url provided.... if not it will conjoin it with the base url

	Parameters
	----------
	base_url:str
		The base url eg . https://site.com
	url:str
		The url in question

	Returns
	-------
	str
		A url that meets requirements

	'''
	
	if url.startswith("http"):
		return url
	new_url =urllib.parse.urljoin(base_url , url)
	return new_url



import bs4 
from bs4 import BeautifulSoup 
from bs4.element import AttributeValueList


def get_first_with_attribute( soup:BeautifulSoup , tag_name:str , attribute:str , empty_attribute:bool = True  ):
	''' This function tries to return the  first matching element in a soup object that has the attribute present 


	Parameters
	----------
	soup:BeautifulSoup
		The BeautifulSoup instance containing the elements to be searched
	tag_name:str
		The tag name to match 
	empty_attribute:bool
		Whether or not the attribute being searched should contain a value . ie. It is not empty 

	Returns
	-------
	tag:bs4.Tag 
		The element found 
	None
		If no matching element is found
	'''


	tags = soup.find_all(tag_name)

	for tag in tags:
		if tag.has_attr(attribute):
			value =  tag.attrs[attribute ]
			if not  empty_attribute:
				if value.strip() != "":
					continue 
			return tag 

	return None 


	


import os 
import csv
import json 
import openpyxl
from collections.abc import Iterable
from openpyxl import Workbook
from crawlfish.crawlfish_utils import check_file
from .save_options import SaveOption , InvalidSaveOptionError, CSVSaveOption , ExcelSaveOption , JSONSaveOption,JSONFileSaveOption



class ListDataSaver:
	'''
	This class contains methods for saving scraped data in different file formats

	'''
	@staticmethod
	def save_by_option(data:Iterable , save_option:SaveOption):
		'''Saves data by the specified option

		Parameters
		----------
		data:Iterable
			The data to e saved 
		save_option:SaveOption
			The save option instance to use 

		Returns
		-------
		data:[dict,list]
			The data saved
		'''
		if not isinstance(save_option , SaveOption):
			raise InvalidSaveOptionError( save_option)

		# CSV
		if isinstance(save_option ,CSVSaveOption ):
			return ListDataSaver.to_csv_file( data ,save_option.file , 
											save_option.overwrite , save_option.delimiter )
		# Excel spreadsheet
		if isinstance(save_option , ExcelSaveOption ):
			return ListDataSaver.to_spreadsheet_file(data , save_option.file , 
													save_option.sheet_name , save_option.overwrite)
		# JSON object 
		if isinstance(save_option , JSONSaveOption ):
			return ListDataSaver.to_json(data , save_option.key_index)
		# JSON file
		if isinstance(save_option , JSONFileSaveOption ):
			return ListDataSaver.to_json_file(data , save_option.key_index , save_option.file ,  save_option.overwrite ,
											 save_option.ensure_ascii  , save_option.indent  , 
											 save_option.encoding  )
		raise InvalidSaveOptionError(save_option)

	@staticmethod
	def to_csv_file(  data:Iterable , file:str = "Scraped_data.csv" , 
						overwrite:bool = True , delimiter:str="," ):
		'''Saves data to  CSV file 

		Parameters
		----------
		data:Iterable
			The data to be saved
		file:str
			The file where the data is to be stored
		overwrite:bool
			Whether or not to overwrite existing data in the file
		delimiter:str
			The delimiter in the csv file 


		Returns
		-------
		data 

		'''
		data = data 
		original_data = data
		data_to_write = []
		if not check_file(file):
			raise FileNotFoundError("Could not write to file - {}".format(file))
		# try and get existing data
		file_size = os.path.getsize(file)
		existing_data = []
		if file_size > 0:
			# check for existing data
			with open(file, 'r' ) as f:
				existing_data = [ row for row in csv.reader(f)]
		if overwrite:
			data_to_write = data
		else:
			data_to_write = existing_data + data[1:] 

		with open(file , "w" , newline= '') as f:
			writer = csv.writer(f , delimiter = delimiter)
			for row in data_to_write:
				writer.writerow(row)
		return original_data


	@staticmethod 
	def to_spreadsheet_file(data:Iterable , file:str , sheet_name = "data" , 
					overwrite = True):
		'''Saves data to an excel spreadheet file . 
		The program will not overwite existing data provide that a unique sheet name id provided
		
		Parameters
		----------
		data:Iterable
			The data to be saved
		file:str
			The file where the data is to be stored
		overwrite:bool
			Whether or not to overwrite existing data in the file
		sheet_name:str
			The name of the sheet in the excel file where the data is to be stored

		Returns
		-------
		data
		'''
		original_data = data 
		data = data 
		if not check_file(file):
			return FileNotFoundError("Could not create or write to file - {}".format(file))
		# try and get existing data 
		workbook = ""
		sheet = ""
		file_size = os.path.getsize(file)

		if file_size > 0:
			# file contains something maybe 
			try:
				workbook = openpyxl.load_workbook(file)
			except:
				raise 
		if not workbook:
			workbook = Workbook()
		if sheet_name in workbook.sheetnames:
			if overwrite:
				del workbook[sheet_name]
				sheet = workbook.create_sheet(sheet_name)
			else:
				# remove the header from data since we are now appending to existing data 
				if len(data) > 1:
					data = data[1:]
				else:
					data = []
				sheet = workbook[sheet_name]
		else:
			sheet = workbook.create_sheet(sheet_name)
		for row in data:
			sheet.append(row)
		workbook.save(file)
		return original_data 


	@staticmethod
	def to_dict(data:Iterable , key_index:int=0):
		'''Converts the data to a dict object
		
		Parameters
		----------
		data:Iterable
			The list to convert to a dict
		key_index:int
			The index of the item in each list in the data list that is to be used as a key 

		Returns
		-------
		data_in_dict_form:dict
			The data in dictionary form

		'''
		data_in_dict_form ={}
		headers = data[0]
		for row in data:
			if data.index(row) == 0:
				continue
			header = row[key_index]
			value = {}
			for item in row:
				index  = row.index(item)
				if index == key_index:
					continue
				print(index , headers)
				value[headers[index]] = item 
			data_in_dict_form[header] = value 
		return data_in_dict_form



	@staticmethod
	def to_json(data:Iterable , key_index = 0 ):
		'''Converts the data to a json object ,same as to_dict , WILL BE DEPRECATED

		Parameters
		----------
		data:Iterable
			The list to convert to a dict
		key_index:int
			The index of the item in each list in the data list that is to be used as a key 

		Returns
		-------
		data_in_dict_form:dict
			The data in dictionary form'''
		data_in_dict_form = ListDataSaver.to_dict(data , key_index = key_index)
		 
		return data_in_dict_form

	@staticmethod
	def to_json_file(data:Iterable ,
				 key_index = 0,
				  file:str = "scraped_data.json" ,
				 overwrite:bool = True,
				  ensure_ascii:bool = False  ,
				   indent:int = 4 ,
				    encoding:str = 'utf-8' ):
		'''Converts the data to a json object

		Parameters
		----------
		data:Iterable
			The list to convert to a dict
		key_index:int
			The index of the item in each list in the data list that is to be used as a key 
		file:str
			The name of the JSON file where the data will e stored
		overwrite:bool
			Whether or not to overwrite existing data in the file
		ensure_ascii:bool
			Whether or not to ensure ascii characters
		indent:int
			The indentation to be used in the file
		encoding:str
			The encoding to e used during writing to the file 
			
		Returns
		-------
		data_in_dict_form:dict
			The data in dictionary form
		'''
		if not check_file(file):
			raise FileNotFoundError("Could not load or write to file :( ")
		data_in_dict_form = ListDataSaver.to_dict(data , key_index = key_index)
		existing_json = ""
		try:
			with open(file , "r")  as f:
				existing_json = json.load(f)
		except (UnicodeDecodeError , json.JSONDecodeError) as e :
			pass 
		except:
			raise
		data_to_write = {}
		if overwrite:
			data_to_write = data_in_dict_form
		else:
			data_to_write = existing_json | data_in_dict_form
		with open(file, 'w', encoding=encoding) as f:
			json.dump(data_to_write, f, ensure_ascii=ensure_ascii, indent=indent)
		return data_in_dict_form
	




from .http import *
from .html import *
from .crawler import *
from .crawlfish_utils import *
from crawlfish.data_saving_formats import save_formats



__all__ = [
		
	'get_url' , 
	'get_page' , 
	'get_page_with_driver',
	"ensure_soup", 
	"ElementFinder" , 
	"get_elements_text",
	"get_element_text",
	"get_first_text_value", 
	"get_first_attr_value",
	"check_file" , 'check_folder',
	'download_static_code_file',
	'download_static_code_files',
	'download_video',
	'download_videos',
	'download_image',
	'download_images',
	'explore_webpage',
	'explore_website',
	'extract_file_info_from_url',
	'ensure_html_url',
	'UrlsCrawler',
	'ListDataSaver',
	'WebsiteReport',
	'WebpageReport',
	'ExcelSaveOption',
	'JSONSaveOption',
	'JSONFileSaveOption',
	'CSVSaveOption',
	'save_formats'
]


import bs4 
from collections.abc import Iterable
from bs4 import BeautifulSoup
from .get_element_text import get_element_text
from .element_finder import ElementFinder

def get_first_attr_value( 
	html_or_soup:tuple[str,BeautifulSoup]   ,
	tag_name:str ,
	attribute:str ,
	tag_names:Iterable  = [],  
	attributes:Iterable = [] , 
	text_match:str = '' ,
	text_matches:Iterable = []   ):
	''' This functions returns the first non-empty attribute value from elements that match the search parameters 
	
	Parameters
	-----------
	html_or_soup:BeautifulSoup
		The html string or  soup to search from
	tag_name:str
			tag name to match
	tag_names:abc.Iterable
		Tag names to match 
	attribute:str
		Element attribute name to match
	attributes:list
			A list of attribute names to match 
	text_match:str
		The text value to match  , it can also be a regex 
	text_matches:list
		A list of text valies to match against element text values . They can also be regex
	 
	Returns
	-------
	value:str
		The first non-empty attribute value found 
	'''
	tag_names = tag_names + [tag_name,]
	attributes = attributes + [attribute,]
	text_matches = text_matches + [text_match,]
	finder = ElementFinder(html_or_soup)
	matching_elements  = finder.find_elements(tag_names = tag_names, attributes = attributes) + finder.find_elements(attributes = attributes,  text_matches = text_matches)
	for element in matching_elements:
		for attribute in attributes:
			value =element[attribute].strip() 
			if value  !='' :
				return value 

	return ''



import requests 
from crawlfish import explore_website ,explore_webpage , WebpageReport , WebsiteReport


webpage_url = "https://www.sample-videos.com/"

def custom_get_function(url):
	'''A custom fetch function to be used by crawlfish 
	It returns a requets.Response object'''
	print("Custom getting website")
	response = requests.get(url)
	print("Foun response")
	return response


# explore a website
website_report = explore_website(webpage_url , get_function = custom_get_function , crawl_limit = 2)


# explore a web page
webpage_report = explore_webpage(webpage_url , get_function = custom_get_function)


# Saving a webpage_report 
webpage_report.save(output_folder = "test_data/mypagereport/"  )
website_report.save(output_folder = "test_data/mysitereport/"  )

# You can also change the function any time and when you use
# the save()  method next time , it will fecth urls using the provided functions
webpage_report.get_function = custom_get_function
website_report.get_function = custom_get_function






from crawlfish import ListDataSaver

# This is just a class  that may come in handy  when you want to save data 
# that is in a list 

data = [
	["Mike" , "alburqueque" , "herewhatsgonnahappen@gmail.com",],
	["Jesse" , "alburqueque" , "thebomb_yo#gmail.com"],
	["Tuco Salamanca" , "Mexico" , "youpunkinme@gmail.com"],
	]

headers = [ "name" , "address" , "email" ]
saver = ListDataSaver()
# remember to add you headers at the start of the data list 
data = [headers,] + data 

# save to csv
saver.to_csv_file(data ,   "test_data/myfile.csv" ,overwrite = False)

# save to spreadsheet
saver.to_spreadsheet_file( data , 'test_data/file.xlsx' , overwrite = True )

# save to json 
json_format = saver.to_json(data , key_index = 2)
json_format


# save to json file
file = "test_data/data.json"
saver.to_json_file( data , key_index = 1 , file = file ,overwrite = True , indent = 4 )
with open(file , "r") as f:
	print(f.read())



import bs4 
from collections.abc import Iterable 
from .get_element_text import get_element_text
from .element_finder import ElementFinder

def get_first_text_value( 
	html_or_soup:bs4.BeautifulSoup  ,
	tag_name:str ,
	attribute:str = '' ,
	tag_names:Iterable = [],  
	attributes:Iterable= [] , 
							    ):
	''' This functions returns the first non-empty text value from elements that match the search parameters 
	
	Parameters
	-----------
	html_or_soup:BeautifulSoup
		The html string or  soup to search from
	tag_name:str
			tag name to match
	tag_names:list
		Tag names to match 
	attribute:str
		Element attribute name to match
	attributes:list
			A list of attribute names to match 
	 
	Returns
	-------
	value:str
		The first non-empty text value found 
	empty_string:str
		An empty string if nothing is found 
	'''
	tag_names = tag_names + [tag_name,]
	attributes = attributes + [attribute,]
	finder = ElementFinder(html_or_soup)
	matching_elements  = finder.find_elements(tag_names = tag_names, attributes = attributes) 
	for element in matching_elements:
		value = get_element_text(element).strip()
		if value  !='' :
			return value 
	empty_string = ''
	return empty_string


import requests
from bs4 import BeautifulSoup

def ensure_soup(item:tuple[BeautifulSoup , requests.Response , str]):
	'''
	A little helper function that will try to return a BeautifulSoup object from the item provided
	return the item if its a soup

	Parameters
	----------
	item:[BeautifulSoup , requests.Response , str]
		The item to get a soup instance from 
	'''
	default_soup = BeautifulSoup("" , "html.parser")
	soup = ""
	if type(item ) == BeautifulSoup:
		return item
	if  type(item) == str:
		return BeautifulSoup(item , "html.parser")
	if type(item) == requests.Response:
		return BeautifulSoup(item.content , "html.parser")
	return default_soup




import re , bs4 , os 
from collections.abc import Iterable
from tqdm import tqdm 
from bs4 import BeautifulSoup
from bs4.element import AttributeValueList
from .get_element_text import get_element_text
from .ensure_soup import ensure_soup


class ElementFinder:

	''' 
	This class is used to perform custom element searches within a html string or BeautifulSoup object
	
	Parameters
	----------
	html_or_soup:[str,BeautifulSoup]
		The html info to search from
	'''
	def __init__(
		self ,
		html_or_soup:tuple[str,BeautifulSoup] 
		):
		self.soup = ensure_soup(html_or_soup)
		self.html = self.soup.prettify()

	def find_elements(  
		self , 
		tag_name:str="" ,
		tag_names:Iterable=[] ,
		attribute:str = "" ,
		attributes:Iterable = [] ,
		attribute_value:str="" ,
		attribute_values:Iterable = [] , 
		text_match:str= "" ,
		text_matches:Iterable = [] , 
		show_progress:bool = True ,
		num:int = 0
						    ):
		'''
		This function returns a list if elements that match all the provided specifications
		  and returns an empty list if none is found 

		Parameters
		----------
		tag_name:str
			tag name to match
		tag_names:Iterable
			Tag names to match 
		attribute:str
			Element attribute name to match
		attributes:Iterable
			A list of attribute names to match 
		attribute_value:str
			The attribute value to match , it can also be a regex
		attribute_values:Iterable
			A list attribute values to match , they can also be a regex
		text_match:str
			The text value to match  , it can also be a regex 
		text_matches:Iterable
			A list of text valies to match against element text values . They can also be regex
		show_progress:bool
			A flag to control verbosity, has not been inplemented yet , so no need using this now
		num:int
			The maximum number of elements to find . num = 0 returns all matched elements
		

		Returns
		-------
		elements:list 
			A list of elements that match the specified requirements 
			The items in the list are of type bs4.Tag

		'''
		print("Searching for elements ")
		soup = self.soup  
		tag_names = tag_names + [tag_name,]
		attributes = attributes + [attribute,]
		attribute_values = attribute_values + [ attribute_value ,]
		text_matches = text_matches + [text_match,]
		elements_that_match_tag_names = []
		elements_that_match_attributes = [] 
		elements_that_match_attribute_values = [] 
		elements_that_match_text_matches = []
		all_elements = soup.find_all()
		# # collect elements_that_match_tag_names
		# for tag_name in tag_names:
		# 	elements = soup.find_all(tag_name)
		# 	elements_that_match_tag_names += elements


		what_to_match = {
				"tag_name":True if ''.join(tag_names).strip() != '' else False  , 
				'attribute':True if ''.join(attributes).strip() != '' else False,
				'attribute_value':True if ''.join(attribute_values).strip() !=  '' else False,
				'text':True if ''.join(text_matches).strip() !=  '' else False,

		}
		matched_elements = [] 
		for element in tqdm(all_elements  , desc = "Finding elements "  ):
			if element in matched_elements:
				continue
			if not type(element) == bs4.Tag:
				continue
			tag_name_check = False 
			attribute_check = False
			attribute_value_check = False
			text_check  = False
			all_checks = []
			# check tag name 
			if what_to_match['tag_name']:
				if element.name in tag_names:
					tag_name_check = True
					all_checks.append(tag_name_check)
				else:
					all_checks.append(False)
			# check attribute 
			if what_to_match['attribute']:
				for attr in list(element.attrs.keys()):
					for attr_to_check in attributes:
						if attr == attr_to_check:
							attribute_check = True
							break 
				all_checks.append(attribute_check) 
							
			# check attriubte values
			if what_to_match['attribute_value']:
				for attr_value in list(element.attrs.values()):
					if type(attr_value) == AttributeValueList:
						if len(attr_value) > 0:
							attr_value = attr_value[0]
						else:
							continue
					if attr_value == '':
						continue
					for attr_value_to_check in attribute_values:
						if re.search(attr_value_to_check , attr_value ):
							attribute_value_check = True
							break
				all_checks.append(attribute_value_check)

			# check text matches
			if what_to_match['text']:
				for  text_to_match in text_matches:
					if text_to_match in (None , ''):
						continue
					element_text = get_element_text(element)
					if element_text.strip == "":
						continue
					if re.search(text_to_match , element_text):
						text_check = True 
						break 
				all_checks.append(text_check)

			# tag_name_check if what_to_match['tag_name'] else False,
			# attribute_check if what_to_match['attribute'] else False ,
			# attribute_value_check if what_to_match['attribute_value'] else False ,
			# text_check if what_to_match['text'] else False , ]
			# print("\n ~~ All checks for ", element , all_checks)
			if len(all_checks) == 0:
				continue
			if all(all_checks):
				matched_elements.append(element)
				if num > 0:
					if len(matched_elements ) == num:
						return matched_elements
		print("\n ============ SEARCH COMPLETE ! Found {} elements that match the search =========== \n ".format(len(matched_elements)))
		print(matched_elements)
		return matched_elements




 


import tldextract
from urllib.parse import urlparse   
from string import ascii_uppercase, digits ,ascii_lowercase , printable 


def clean_split_part(part:str):
	'''Remove unwanted url charchters from a url part 

	Parameters
	----------
	part:str
		The part of the url to clean

	Returns
	-------
	result:str
		A string that meets all the requirements 

		 '''
	wanted_charachters = ascii_lowercase + ascii_uppercase + digits  + '_'
	part = part.replace('-', '_')
	result = ''
	for char in part:
		if char in wanted_charachters:
			result += char
	return result


	
def url_to_folder_name(url:str):
	"""Converts a url to a value that can be used as a folder name safely

	Parameters
	----------
	url:str
		The url to get a folder name from 

	Returns
	-------
	folder_name:str
		A folder name that works (is valid or won't raise an error)"""
	# get base url first
	result = tldextract.extract(url)
	domain = result.domain
	path = urlparse(url).path.replace('.', '_')
	split = path.split('/')
	path = "".join(['_'+ clean_split_part(item) for item in split ])
	folder_name = clean_split_part( domain  + path)

	return folder_name



import os 


def check_file(file_path:str , make_if_none:bool = True  ,encoding:str = "utf-8"):
	'''
	 check if the path exists and if specified ,  attempts to create it and return True 

	Parameters
	-----------
	file_path:str
		The file path to check
	make_if_none:bool
		Whether or not to try and make the file if it doesn't exist
	encoding:str
		The encodng to use when attempting to create the file 

	Returns
	-------
	bool
		Whether or not the file exists
	'''
	folder  , file = os.path.split(file_path)
	if not make_if_none:

		return os.path.isfile(file_path)
	# print("Will make if none")
	if os.path.isfile(file_path):
		return True 
	if not file:
		return False
	if not os.path.isdir(folder) and folder.strip() != ''  :
		try:
			os.makedirs(folder)
		except Exception as e:
			print("Could not make file folder "  , 'because of error ', e)
			return False
	print("Writing to file " , file_path)
	try:
		with open(file_path , "w" , encoding = encoding) as f:
			f.write("")
	except (ValueError ,OSError) as e:
		return False
	except Exception as e:
		raise
	return os.path.isfile(file_path)


import json 
import os
from types import FunctionType
from collections.abc import Iterable
from urllib.parse import urlsplit, urlunsplit
from datetime import datetime
from configparser import ConfigParser
import bs4
import tldextract
from bs4 import BeautifulSoup
from tqdm import tqdm
from tabulate import tabulate 
from crawlfish.crawler.download_image import download_image 
from crawlfish.crawler.download_static_code_file import download_static_code_file
from crawlfish.http import get_url
from crawlfish.crawlfish_utils import check_file ,check_folder
from crawlfish.html  import ensure_soup , get_element_text
from .get_base_url import get_base_url
from .get_url_parts import get_url_parts
from .url_to_folder_name import url_to_folder_name
from .extract_file_info_from_url import extract_file_info_from_url
 

PAGE_REPORT_STRING = """
======================================================================================
		  WEB PAGE REPORT
		 `````````````````
{}
======================================================================================
		
	Metadata 
{}

	Urls stats
{}

	Static data
{}

============================================

"""

DATA_NOT_FOUND = "NULL"

 

class WebpageReport:
	'''
	To be used as a container for data obtained after exloring one webpage
	Contains mostly links  and metadata 

	Parameters
	----------
	url:str
		The url of the webpage 
	title:str
		The title of the webpage 
	css_urls:Iterable 
		A list containing urls to css files
	js_urls:Iterable 
		A list containing urls to Javascript files 
	image_urls:Iterable
		A list containing urls to image files 
	css_tags:Iterable 
		A list containing css bs4.Tag  instances that represent <style> tags
	js_tags:Iterable 
		A list containing  bs4.Tag  instances that represent Javascript <script> tags
	json_tags:Iterable
		A list containing  bs4.Tag  instances that represent JSON <script> tags
	metadata:dict 
		A dictionary object containing the page's metadata info { key:value ... }
	other_site_urls:Iterable
		A list of urls to other websites 
	same_site_urls:Iterable 
		A list of urls found on the webpage that lead to pages of the same website
	soup:BeautifulSoup 
		A bs4.BeautifulSoup instance of the webpage's code 
	get_function:FunctionType
		A function that takes a url as an argument and returns a request.Response object 
	html:str
		The html code of the webpage 

	'''
	def __init__(
		self ,
		url:str ,
		title:str =DATA_NOT_FOUND,
		css_urls:Iterable = [] ,
		js_urls:Iterable = [] , 
		image_urls:Iterable=[],
		css_tags:Iterable = [] ,
		js_tags :Iterable= [] ,
		json_tags:Iterable = [],
		metadata:dict = {},
		other_site_urls:Iterable=[] , 
		same_site_urls:Iterable = [] ,
		soup:BeautifulSoup =None, 
		get_function:FunctionType = get_url,
  		html ='' 
  		):
		tld_result = tldextract.extract(url)
		self.domain = tld_result.domain
		self.url = url
		self.base_url  = get_base_url(url)
		self.soup = soup
		self.get_function = get_function
		self.html = ''
		self.elements_processed = 0
		if soup:
			self.html = soup.prettify()
			self.elements_processed = len(soup.find_all())
		else:
			self.soup = bs4.BeautifulSoup('' ,'html.parser')
		self.css_urls  = css_urls
		self.js_urls = js_urls
		self.image_urls = image_urls
		self.css_tags = css_tags 
		self.js_tags = js_tags 
		self.json_tags = json_tags
		self.title = title 
		self.metadata = metadata
		self.same_site_urls = same_site_urls
		self.split_url = urlsplit(url)
		self.network_location = self.split_url.netloc
		self.discovered_urls = []
		self.all_site_urls = []
		self.other_site_urls = other_site_urls
		self.internal_css_code ={}
		self.external_css_code ={}
		self.internal_js_code ={}
		self.external_js_code = {} 
		self.image_data = {}
		self.internal_json_code = {}
		self.already_downloaded_static = {}
		self.all_static_urls = self.image_urls + self.js_urls + self.css_urls
		self.all_static_data = {} # all static data . images, js , css
		self.DiscoverMoreSameSiteLinks()
		self.ExtractInternalCode()


	def DiscoverMoreSameSiteLinks(self):
		'''Looks for possible undiscovered urls of the same website '''
		discovered_urls = self.discovered_urls
		for url in self.same_site_urls:
			parts = get_url_parts(url)
			for i in parts :
				if not i in self.same_site_urls and not i in discovered_urls:
					discovered_urls.append(i)
		self.discovered_urls =  discovered_urls  
		print("Discovered {} possible same site  links ".format(len(self.discovered_urls)))
		self.all_site_urls = self.discovered_urls + self.same_site_urls

	def as_text(self):
		'''Returns this class' data in text form '''
		return self.__repr__()

	def as_dict(self  ):
		''' Returns the page data info as a dict object 

		'''
		data = {

				"static_stats":{ 
					'js_tags':len(self.js_tags),
					'css_tags':len(self.css_tags),
					'js_urls':len(self.js_urls),
					'css_urls':len(self.css_urls),
					'image_urls':len(self.image_urls),
					'json_tags':len(self.json_tags),
					"elements_processed":self.elements_processed,
						},
				"metadata":self.metadata,
				"urls":{
					"base_url":self.base_url,
					"network_location":self.network_location,
					"same_site_links":len(self.same_site_urls),
					"discovered_urls":len(self.discovered_urls),
					"other_site_urls":len(self.other_site_urls),
				},

				}

			 
		
		return data 

	def save_to_config_file(self , file:str ,encoding:str='utf-8'):
		''' Save the page report data in config  file using configparser
		Parameters
		----------
		file:str
			The name of the file where the data will be saved
		encoding:str
			The encoding to use during file read and write 

		Returns
		-------
		None 
		'''
		print("Saving Web page report to CONFIG file {} -------- ".format(file) , end = '')
		check_file(file)
		cp =ConfigParser()
		as_dict = self.as_dict()

		for title,data in as_dict.items():
			cp[title] = {}
			for k , v in data.items():
				cp[title][k] = str(v)
		with open(file , 'w' ,encoding =encoding) as f:
			cp.write(f)
		print("DONE !")



	def save_to_json_file(self , file:str, ensure_ascii:bool = False  ,
					 indent:int = 4 , encoding:str = "utf-8"):
		''' Saves this class' information in to a json file 

		Parameters
		----------
		file:str
			The name of the file where the data will be saved
		ensure_ascii:bool
			Ensures ascii in the json file 
		indent:int
			The indentation to use in the JSON file 
		encoding:str
			The encoding to use during file read and write 
		
		Returns
		-------
		None 
		'''
		print("Saving Web page report to JSON file {} -------- ".format(file) , end = '')
		to_write = self.as_dict()
		with open(file ,"w" ,encoding = encoding) as f:
			json.dump(to_write ,f  ,ensure_ascii=ensure_ascii, indent=indent)

		print("DONE")


	def save_to_text_file(self, file:str ,encoding:str = 'utf-8' ):
		''' Saves this class' information in to a text file

		Parameters
		----------
		file:str
			The name of the file where the data will be saved
		encoding:str
			The encoding to use during file read and write 
		
		Returns
		-------
		None 
		'''
		print("Saving Web page report to TEXT file {} -------- ".format(file) , end = '')
		check_file(file)
		data = self.as_text()
		with open(file , "w" ,encoding = encoding) as f:
			f.write(data)
		print("DONE")



	def ExtractInternalCode(self):
		''' Extract the code from css and js related tags '''
		print("Extracting JS internal code --- " , end ='')
		for tag in self.js_tags:
			code = get_element_text(tag)
			tag_id = "script_tag_{}.js".format(self.js_tags.index(tag))
			self.internal_js_code[tag_id] = code 
		print("DONE")
		print("Extracting CSS internal code --- " , end ='')
		for tag in self.css_tags:
			code = get_element_text(tag)
			tag_id = "script_tag_{}.css".format(self.css_tags.index(tag))
			self.internal_css_code[tag_id] = code 
		print("DONE")
		print("Extracting JSON code ------------" , end = '')
		for tag in self.json_tags:
			code = get_element_text(tag)
			tag_id = "script_tag_{}.json".format(self.json_tags.index(tag))
			self.internal_json_code[tag_id] = code 
		print("DONE")

	def FetchExternalCode(self , to_fetch = ["js" , "css"]):
		'''Download and store all external static  code

		Parameters
		----------
		to_fetch:Iterable 
			A list of the static file types to fetch 

		Returns
		-------
		None 

		'''
		if 'js' in to_fetch:
			self.FetchExternalJS()
		if 'css' in to_fetch:
			self.FetchExternalCSS()
		
	def FetchExternalCSS(self):
		"""Fetches external CSS code """
		for url in tqdm(self.css_urls ,"Fetching external CSS--"):
			file_name =extract_file_info_from_url(url)[0] 
			if file_name in self.already_downloaded_static:
				# dont download 
				# add the data 
				self.external_css_code[file_name] = self.already_downloaded_static[file_name]
				continue
			content ,base_name = download_static_code_file( url  , self.get_function , save_to_file = False)
			if not content.strip():
				continue
			self.external_css_code[base_name] = content

	def FetchExternalJS(self):
		"""Fetches external JS code"""
		for url in tqdm(self.js_urls ,"Fetching external Javascript--"):
			file_name =extract_file_info_from_url(url)[0] 
			if file_name in self.already_downloaded_static:
				# dont download 
				# add the data 
				self.external_js_code[file_name] = self.already_downloaded_static[file_name]
				continue
			base_name , content= download_static_code_file( url  , self.get_function ,save_to_file = False)
			if not content.strip():
				continue
			self.external_js_code[base_name] = content



	def FetchImages(self):
		''' Gets and saves the images in the image_utls attribute 

		Returns
		-------
		None

		'''
		for url in tqdm(self.image_urls , "Fetching {} images --- ".format(len(self.image_urls))):
			file_name =extract_file_info_from_url(url)[0] 
			if file_name in self.already_downloaded_static:
				# dont download 
				# add the data 
				self.image_data[file_name] = self.already_downloaded_static[file_name]
				continue
			base_name ,data = download_image(url , self.get_function , save_to_file = False)
			self.image_data[base_name] = data

	def save_images(self , output_folder:str = os.getcwd()):
		'''Saves the downloaded image data to files in the specified output folder 

		Parameters
		----------
		output_folder:str
			The folder where to save the image files 

		Returns
		-------
		None

		'''
		for base_name , data in tqdm(self.image_data.items() , "Downloading {} image(s)".format(len(self.image_data.keys()))):
			file = os.path.join(output_folder , 'images/' + base_name)
			if not check_file(file):continue
			with open(file , 'wb') as f:
				f.write(data)

	def save_static_code(self , file_code_dict:dict ,
						encoding:str ='utf-8' ,
						tqdm_message:str = "Saving static code"):
		'''Saves static code in the file_code_dixt dictionary into files

		Parameters
		----------
		file_code_dict:dict
			The file and data information in the format { filename:data ..... } 
		encoding:str
			The encoding to use during file read and write 
		tqdm_message:str
			The progress bar message	

		Returns
		-------
		None 

		'''
		for file ,code in tqdm(file_code_dict.items() , tqdm_message):
			if not check_file(file):
				continue
			with open(file, "w" , encoding=encoding) as f:
				f.write(code)

	def save_js_code(self , output_folder = os.getcwd() ,encoding = "utf-8"):
		''' Save javascript into files

		Parameters
		----------
		encoding:str
			The encoding to use during file read and write
		output_folder:str
			The folder where to save the JS files   

		Returns
		-------
		None 

		'''
		external_code_folder = os.path.join(output_folder ,"js/external")
		internal_code_folder = os.path.join(output_folder ,"js/internal")
		external_file_code_dict = {  os.path.join(external_code_folder , base_name):code for base_name,code in self.external_js_code.items()  }
		internal_file_code_dict = {  os.path.join(internal_code_folder , base_name):code for base_name,code in self.internal_js_code.items()  }
		self.save_static_code(external_file_code_dict , tqdm_message = "Saving external JS code ")
		self.save_static_code(internal_file_code_dict, tqdm_message = "Saving internal JS code ")

	def save_css_code(self , output_folder = os.getcwd() ,encoding = "utf-8"):
		''' Save CSS CODE into files 

		Parameters
		----------
		encoding:str
			The encoding to use during file read and write
		output_folder:str
			The folder where to save the CSS files   

		Returns
		-------
		None '''
		external_code_folder = os.path.join(output_folder ,"css/external")
		internal_code_folder = os.path.join(output_folder ,"css/internal")
		external_file_code_dict = {  os.path.join(external_code_folder , base_name):code for base_name,code in self.external_css_code.items()  }
		internal_file_code_dict = {  os.path.join(internal_code_folder , base_name):code for base_name,code in self.internal_css_code.items()  }
		self.save_static_code(external_file_code_dict , tqdm_message = "Saving external CSS code ")
		self.save_static_code(internal_file_code_dict, tqdm_message = "Saving internal CSS code ")

	def save_json_code(self , output_folder = os.getcwd() ,encoding = "utf-8"):
		''' Save CSS CODE into files 
		Parameters
		----------
		encoding:str
			The encoding to use during file read and write
		output_folder:str
			The folder where to save the JS files   

		Returns
		-------
		None '''
		internal_code_folder = os.path.join(output_folder ,"json/internal")
		internal_file_code_dict = {  os.path.join(internal_code_folder , base_name):code for base_name,code in self.internal_json_code.items()  }
		self.save_static_code(internal_file_code_dict, tqdm_message = "Saving internal JSON code ")

	


	def save(self  , output_folder:str = os.getcwd()  , download_images:bool = True , download_js:bool =True,
			download_css:bool = True , name_space:bool = True ) :
		'''Save a everything in ideal formats to a  specified outfolder 

			Structure example:

			---- Output folder/
					---- Report.txt
					---- Report.config
					---- Report.json
					---- page.html
					---- static/
							---- js/
							---- css/
							---- images/
							---- json 


		Parameters
		----------
		output_folder:str
			The folder where to save the JS files 
		download_images:bool
			Whether or not to download images
		download_js:bool
			Whether or not to download external Javascript static files  
		download_css:bool
			Whether or not to download external CSS static files  
		name_space:bool
			Whether or not to uniquely namespace the output folder using a time tag eg. OutputFolder_3rd_March/
		
		Returns
		-------
		None

		'''

		# reports
		if name_space:
			now = datetime.now()
			string_time = now.strftime("%H_%M_%A_%d_%Y")
			name = self.domain +"_WebpageReport_"+string_time
			output_folder = os.path.join(output_folder, name)
		check_folder(output_folder)
		json_report_file = os.path.join(output_folder , "Report.json")
		text_report_file = os.path.join(output_folder , "Report.txt")
		config_report_file = os.path.join(output_folder , "Report.config")
		html_file = os.path.join(output_folder , url_to_folder_name(self.url) + "_html_content.html")
		print("Writing report files ------------- " ,end ='')
		self.save_to_text_file(text_report_file)
		self.save_to_config_file(config_report_file)
		self.save_to_json_file(json_report_file)
		print("DONE!")
		print("Writing same site links info ----- " , end = '')
		with open(os.path.join(output_folder , "same_site_urls.txt") , 'w') as f:
			for l in self.same_site_urls:
				f.write(l + '\n')
		print("DONE")
		print("Writing other site links info ----- " , end = '')
		with open(os.path.join(output_folder , "other_site_urls.txt") , 'w') as f:
			for l in self.other_site_urls:
				f.write(l + '\n')
		print("DONE")
		#STATICS
		print("Loading static resources location  information  -----", end= "")
		if download_js:
			self.FetchExternalJS()
		if download_css:
			self.FetchExternalCSS()
		if download_images:
			self.FetchImages()
		print("External static code and resources fetched successfully .. ")
		print("saving static code ")
		self.save_js_code(output_folder = output_folder)
		self.save_css_code(output_folder = output_folder)
		self.save_images(output_folder = output_folder)
		self.save_json_code(output_folder = output_folder)
		print("Static code saved .")
		self.all_static_data.update(self.external_css_code )
		self.all_static_data.update( self.external_css_code )
		self.all_static_data.update( self.image_data)
		print("All static data parsed.")
		print("Writing HTML file ")
		with open(html_file , "w" , encoding = 'utf-8') as f:
			f.write(self.soup.prettify())
	


	def __repr__(self):
		# meta table
		metadata_table_list = [ [k,v] for k,v in self.metadata.items() ]
		metadata_table = tabulate(metadata_table_list)
		same_site_urls_number =len(self.same_site_urls)
		js_urls_number = len(self.js_urls)
		css_tags_number = len(self.css_tags)
		js_tags_number = len(self.js_tags)
		css_urls_number = len(self.css_urls)
		json_tags_number = len(self.json_tags)
		discovered_urls_number = len(self.discovered_urls)
		# urls stats
		urls_stats_table_list = [
			["Links from same site",same_site_urls_number] ,
			["Other site urls ", len(self.other_site_urls)],
			["Discovered links", discovered_urls_number],
			["Base url", self.base_url],
			["Network location", self.network_location],
		]
		url_stats_table = tabulate(urls_stats_table_list)
		# Static data 
		static_data_stats_table_list = [
			["External Javasript files", js_urls_number],
			['External CSS files' , css_urls_number],
			['Javascript  <script> tags',js_tags_number],
			["JSON tags " , json_tags_number],
			["CSS <style> tags" ,css_tags_number],
			["Image files", len(self.image_urls),], 
			["HTML tags processed" , self.elements_processed],
			]
		static_data_stats_table = tabulate(static_data_stats_table_list)
		repr_str = PAGE_REPORT_STRING.format(
			self.url,
				metadata_table,
				url_stats_table,
				static_data_stats_table)
		return repr_str
 

					 
	def __str__(self):
		return self.__repr__()


import os
from unittest import TestCase
from selenium import webdriver
from selenium.webdriver.firefox.options import Options

from crawlfish import get_page_with_driver


options = Options()
options.add_argument("--headless")
print("Creating a Firefox driver for testing ..........." , end = '')
test_driver = webdriver.Firefox( options = options)
print("DONE !")
test_url = "https://www.sample-videos.com/"
test_output_file = "test_data/Test_get_page_with_webdriver/samplevides.html"


class Test_get_page_with_webdriver(TestCase):

	def test_Raises_driver_type_error(self):
		with self.assertRaises(ValueError):
			get_page_with_driver(test_url , 'driver')

	def test_get_html(self):	
		page = get_page_with_driver(test_url , test_driver)
		self.assertTrue(page.strip() != '')

	def test_save_to_file(self):
		if os.path.isfile(test_output_file):
			os.remove(test_output_file)
		page = get_page_with_driver(test_url , test_driver , output_file = test_output_file)
		self.assertTrue(os.path.isfile(test_output_file))
		with open(test_output_file , "r") as f :
			data = f.read()
			self.assertTrue(len(data) > 0)


from crawlfish import get_page , get_url , get_page_with_driver

# get a requests.Response object of a page
response = get_page("https://www.sample-videos.com")

# get the html of a page
response = get_page("https://www.sample-videos.com")

# get the bs4.BeautifulSoup instance fo a page
soup = get_page("https://www.sample-videos.com" , soup = True )

# save the page to a file 
response = get_page("https://www.sample-videos.com" , output_file = "test_data/stack.html")

# Fetch a page with a  selenium webdriver
from selenium import webdriver 
from selenium.webdriver.firefox.options import Options
# Initialize your webdriver
options = Options()
options.add_argument("--headless")
driver = webdriver.Firefox(options = options)
html = get_page_with_driver( 'https://www.sample-videos.com' , driver)

# Fetch using a selenium webdriver and save a page to a file 
html = get_page_with_driver( 'https://www.sample-videos.com' , driver , output_file = "test_data/withdriver.html")


driver.quit()



import os
import tldextract
from types import FunctionType
from urllib.parse import urlsplit, urlunsplit
from crawlfish.http import get_url
from crawlfish import ElementFinder , get_element_text
from .get_base_url import get_base_url
from .get_url_parts import get_url_parts
from .webpage_report import WebpageReport
from .website_report import WebsiteReport
from .ensure_url_format import ensure_url_format
from .extract_html_meta import extract_html_meta
from crawlfish.default_values import DATA_NOT_FOUND
from .ensure_html_url import ensure_html_url




def explore_webpage(  
	url:str  ,
	get_function:FunctionType = get_url
	 ):
	'''Crawls and goes through a webpage and returns a WebpageReport instance

	Parameters
	----------
	url:str
		The url of the webpage to be crawled
	get_function:FunctionType
		A function for fetching the webpage ,
		it should take a url as its sole argument and return a html string or BeautifulSoup instance

	Returns
	-------
	WebpageReport
		A WebpageReport instance containing all the juicy information
	'''
	print("\n\n=======================\n\n Exploring page ", url)
	html_or_soup = get_function(url)
	finder = ElementFinder(html_or_soup)
	soup = finder.soup
	base_url = get_base_url(url)
	tld_result = tldextract.extract(url)
	domain = tld_result.domain
	split_url = urlsplit(base_url) 
	print("Found base url " , base_url)
	print("Fetching elements --- " , end = '')
	all_elements = soup.find_all() 
	print("Done")
	# title 
	title_tag = soup.find("title")
	title = DATA_NOT_FOUND
	if not title_tag:
		title = DATA_NOT_FOUND
	else:
		title = get_element_text(title_tag)
	print("Page  <title> tag info fetched ")
	print("Extracting meta data ----- " ,end = '')
	metadata = extract_html_meta(soup)
	print("DONE")
	# Colleting images
	print("Collecting image links ---------------- ")
	image_tags = finder.find_elements(tag_name ="img" , attribute= "src")
	image_urls = [ensure_url_format( base_url , tag['src']) for tag in image_tags if tag["src"].strip() != '']

	print("DONE")
	# JAVASCRIPT DATA
	print("Collecting javascript data................" ,end = '')
	internal_js_tags = finder.find_elements( tag_name= "script"   )
	external_js_tags = finder.find_elements( tag_name= "script"  ,  attribute_value=".js" )
	# print("external_js_tags" , external_js_tags)
	js_files_urls = [ensure_url_format(base_url, tag["src"]) for tag in external_js_tags if tag.has_attr("src")]
	print("DONE")
	# CSS DATA
	print("Fetching CSS data.................." , end = '')
	external_css_tags = finder.find_elements(tag_name = "link" , attributes=["rel",'type'] , attribute_values=["stylesheet",'text/css'])
	css_files_urls = [ensure_url_format(base_url , tag["href"]) for tag in external_css_tags if tag.has_attr("href")]
	internal_css_tags = finder.find_elements(tag_name = "style" )
	print("DONE")
	# JSON DATA
	print("Fetching JSON.............." , end = '')
	json_tags = finder.find_elements(tag_name = 'script' ,attribute = "type" ,  attribute_value = 'application/json')
	print("DONE")
	# Remove JSON tags from inernal js tags
	print("Cleaning internal JS data..................." , end = '')
	json_text = [get_element_text(tag) for tag in json_tags]
	internal_js_tags = [tag for tag in internal_js_tags if not get_element_text(tag) in json_text]
	print("DONE")
	print("Collecting same site  urls ")
	href_tags = finder.find_elements(tag_name= 'a'  , attribute = 'href')
	possible_url_leads = []
	other_site_urls = [] 
	for tag in href_tags:
		link = ensure_url_format(base_url , tag['href'])
		if not ensure_html_url(link):
			print("NOT TRUE URL ")
			continue
		tld_result = tldextract.extract(link)
		if tld_result.domain != domain:
			other_site_urls.append(link)
			continue
		if link in possible_url_leads:
			continue
		possible_url_leads.append(link)
	href_data =  [  tag["href"] for tag in href_tags if base_url in tag["href"] ]
	print("Checking links format")

	# generate the report 
	print("Generating report ---- " , end = '')
	page_report = WebpageReport(

			url = url , 
			title = title , 
			css_urls = css_files_urls , css_tags = internal_css_tags , 
			js_urls = js_files_urls , js_tags = internal_js_tags  , 
			json_tags = json_tags , image_urls = image_urls,
			same_site_urls = possible_url_leads , soup = soup,
			html = finder.html , other_site_urls = other_site_urls, 
			metadata= metadata
		)

	print("Page report generated !")
	print(page_report)
	return page_report




  


import os 
from types import FunctionType
from collections.abc import Iterable
import requests
from tqdm import tqdm 
from crawlfish.http import get_url
from crawlfish.crawlfish_utils import check_file , check_folder
from .extract_file_info_from_url import extract_file_info_from_url

def download_static_code_file(
	url:str ,
	get_function:FunctionType = get_url ,
	file:str = '' ,
	output_folder:str = os.getcwd(),
	encoding:str = 'utf-8' ,
	file_name:str = '' ,
	save_to_file:bool = True
	):

	'''Downloads a static text based file eg "https:fie.js" and saves it to the file path provided if save_to_file =True

	Parameters
	----------
	url:str
		The URL of the file to be  downloaded
	get_function:FunctionType
		A function that takes a url as a parameter and returns a request.Response object 
	file:str
		The name of the file to write the downloaded data to 
	output_folder:str
		The folder/directory where to store the downloaded file 
	save_to_file:bool
		Whether or not to save the data to a file 
	file_name:
		The name of the file withholding the extension name where the downloaded data will be written .
		The extension will be added automatically if this is  provided 

	Returns
	-------
	tuple 
		(file_name,data .....)
'''
	response = get_function(url)
	text = response.text
	url_file_name ,extension = extract_file_info_from_url(url)
	if output_folder:
		output_folder = os.path.abspath(output_folder)
		check_folder(output_folder)
	if output_folder and not file and not file_name:
		file= os.path.join(output_folder , url_file_name )
	if file_name:
		file_name += extension
		file = os.path.join(output_folder , file_name )
	if not file_name:
		file_name = os.path.split(file)[-1]
	if save_to_file:
		print("saving data to file ", file , ' --------- ' , end = '')
		if not check_file(file):
			return file_name ,''
		with open(file , "w" ,encoding = encoding) as f:
			f.write(text)
		print("DONE !")
	return file_name , text

def download_static_code_files(
	urls_list:Iterable,
	get_function = get_url ,
	output_folder = os.getcwd(),
	encoding = 'utf-8'
	):
	'''Downloads text based files from the urls provided in the urls_list

	Parameters
	----------
	urls_list:Iterable 
		An iterable containing the urls of the text-based files to be downloaded
	get_function:FunctionType
		A function that takes a url as an argument and returns a request.Response object 
	output_folder:str
		The folder/directory where to store the downloaded files

	Returns
	-------
	files_data:dict
		{ filename:text_data , ...... }
		A dictionary containing the text data as values and file names as keys
	'''

	files_data = {}
	for url in tqdm(urls_list):
		file_name , data = download_static_code_file(url , get_function = get_function , output_folder  = output_folder, encoding = encoding)
		files_data[file_name] = data
	return files_data


import os 
import sys 
import art
import argparse
from ascii_magic import AsciiArt 
from crawlfish import  crawler , supported
from crawlfish import crawlfish_utils



base_dir = os.path.split(os.path.abspath(__file__))[0]



def exit_program():
	''' For exiting all processes
	'''
	print("-------------- EXITING ---------------")
	sys.exit()


def check_option_urls(urls):
	'''a helper function that ensures the validity of the url(s) provided . Exits the program if a url is invalid'''
	for u in urls :
		if not u.startswith("https"):
			print("Invalid url provided - ", u)
			exit_program()


def get_urls_from_file(file:str):
	print("Collecting urls from file , ", file )
	if not crawlfish_utils.check_file(file , make_if_none = False):
		print("Invalid file path provided . Please make sure that file exists ")
		exit_program()
	urls_in_file = []
	urls = []
	with open(file , "r") as f:
		urls_in_file = f.readlines()
	for u in urls_in_file:
		u = u.strip().replace("\n" , '')
		if u == "":
			continue
		urls.append(u)
	if len(urls) == 0 :
		print("No urls found in file " , file)
		exit_program()
	check_option_urls(urls)
	return urls




logo_image = os.path.join(base_dir , "package_data/images/logo.png")

logo = art.text2art("Crawlfish"  ,font = "stforek")
print(logo)
intro = """

by Victor Kipkemboi | scriptilapia@gmail.com

====================================================================================================================
""".format()
print(intro)



# ARGUMENTS
parser = argparse.ArgumentParser()

work_to_do = parser.add_mutually_exclusive_group( )
work_to_do.add_argument("-d","--download",action = "store_true" , help = "Download a file or files")
work_to_do.add_argument("-cp","--crawl-page",action = "store_true" , help = "Crawl a webpage or webpages")
work_to_do.add_argument("-cs","--crawl-site",action = "store_true" , help = "Crawl a website or websites")

# CRAWLING OPTIONS 
parser.add_argument("-rs", "--redownload-static"  ,action="store_true" , 
			help ="""This option present means that  static files with similar urls from different web pages will be downloaded evertime they are found instead of caching the data since they are most likely similar.  """)

# URL(s) INPUT
parser.add_argument("-u" ,"--url" , help = "The URL(s) you want to process . Make sure to add spaces in between the urls" , nargs = "*" )
parser.add_argument("-uf" ,"--urls-file" , help = "The path or name of a file containing urls to be processed  " )

# FILE FORMATS TO DOWNLOAD
download_formats = parser.add_mutually_exclusive_group()
download_formats.add_argument("-i","--image", action="store_true", help = "Indicate that you are downloading an image. Eg: .png ,  .jpeg , .jpg ")
download_formats.add_argument("-v", "--video", action="store_true", help = "Indicate that you are downloading a video. Eg: .mp4 , .mp3 , .mkv ....")
download_formats.add_argument("-tf", "--text-file",action="store_true",  help="Indicate that you are downloading a text-based file . Eg: .js, .txt , .css")

# STATIC FILES TO NOT DOWNLOAD 
parser.add_argument("-es" , '--exclude-static' ,
					 help = "Static file to not download during crawling processes . This option will have no effect during downloading " , 
					 nargs = "*", 
					 choices = supported.CRAWL_STATIC_TYPES,
					 default = [], 
					)


# OUTPUT DATA STORAGE
output_storage = parser.add_mutually_exclusive_group()
output_storage.add_argument("-f", "--file" , help = """The name to be given to the FILE you want to download. Has no effect when many url(s) are downloaded . Make sure to include the correct extension for convenience""" )
output_storage.add_argument("-fn", "--file-name" ,
				 help = """The name you want to give to the downloaded file . The extension will be added automatically .
				  To be used only when one url has been provided otherwise it has no effect whatsoever"""
				  )
output_storage.add_argument("-outd", "--output-dir" , help= "The directory where you want your output stored . The current directory will be used if none is provided . If the process has no use for this , it will have no effect " )


# LIMITS 
parser.add_argument('-cl' , '--crawl-limit' , 
	help = "The number of maximum urls to crawl when crawling a website or a webpage . 0 = Infinity , the default is 0 . This option takes effect during site crawling only" , 
	type = int , 
	default = 0)

args = parser.parse_args()






print("Starting ------------------------- ")
# prepare output storages
output_folder = os.getcwd()
if args.output_dir:
	output_folder = args.output_dir
urls = []
output_file = None 

if args.url:
	check_option_urls(args.url)
	urls = args.url 
elif args.urls_file:
	urls = get_urls_from_file(args.urls_file)

if urls:
	print("{} url(s) waiting to be processed".format(len(urls)))
else:
	print("NO Urls provided :( ")
	exit_program()
# PROCESS ARGUMENTS 
if args.download:
	print("Downloading {} items ".format(len(urls)) )
	if args.image:
		if len(urls) == 1:
			crawler.download_image(urls[0] , file = args.file , file_name = args.file_name , output_folder = output_folder)
		else:
			crawler.download_images(urls , output_folder = output_folder)
	# download videos(s)
	elif args.video:
		if len(urls) == 1:
			crawler.download_video(urls[0] , file = args.file , file_name = args.file_name , output_folder = output_folder)
		else:
			crawler.download_videos(urls , output_folder = output_folder)
	# download text based files
	elif args.text_file:
		if len(urls) == 1:
			crawler.download_static_code_file(urls[0] , file = args.file , file_name = args.file_name , output_folder = output_folder)
		else:
			crawler.download_static_code_files(urls , output_folder = output_folder)
	print("Process complete !")


elif args.crawl_site:
	print("Crawling {} website(s)".format(len(urls)))
	website_reports= []
	for u in urls :
		report = crawler.explore_website(u , crawl_limit = args.crawl_limit)
		website_reports.append(report)
	for wr in website_reports:
		print("Saving website report " , wr )
		wr.save( output_folder = output_folder , redownload_static = args.redownload_static , exclude_static = args.exclude_static )
	print("Process complete !")

elif args.crawl_page:
	print("Crawling webpage(s)")
	webpage_reports = [] 

	for u in urls :
		page_report  =crawler.explore_webpage(u)
		webpage_reports.append(page_report)
	print("Saving web page reports ----- ")
	download_images  ,download_js , download_css = False if 'img' in args.exclude_static else True  , False if 'js' in args.exclude_static else True  , False if 'css' in args.exclude_static else True  
	for pr in webpage_reports:
		pr.save( output_folder = output_folder , download_css = download_css , download_js = download_js  , download_images = download_images )

	print("Process complete !")


import re 


def get_base_url(url:str):
	''' Gets the main starting point of a website;s url .
	Eg . https//:site.com/page/data/hello 

		 will be :

		 https://site.com
	
	Parameters
	----------
	url:str
		The url to extract the base url from 

	Returns
	-------
	url:str
		The base url found
	'''
	# Look for the third slash 
	needed_charchters = '/' , 'http' , ':' 
	if not all([ char in url for char in needed_charchters  ]):
		print("Could not get base url : {}".format(url))
		return url
	if url.count('/') >= 3:
		occurrences = [ i.start()  for i in re.finditer('/', url)]
		end = occurrences[2] # the third / 
		return url[:end]
	return url





def ensure_html_url(url:str):
	'''Returns a boolean indicating if the url looks like it leads to a html page ,
		ie .The content-type of the request headers will be 'text/html'

	Parameters
	----------
	url:str
		The url to check

	Returns
	-------
	bool
		Whether the url is valid or not 
	'''

	url = url.split('?')[0].lower()
	if not url.startswith('http'):
		return False
	unwanted_ends = ['.jpg' ,'.png' , '.jpeg' , '.jpg','.gif','.svg',

						'.css','.js', '.scss','.json', 
						'.3gp','.mp4a', '.mp4' , 'mp3', '.avi' , '.mkv' ,'.flv' , 
						'.pdf', '.zip' , '.psd' , '.dxf' , '.doc' , '.docx','.ppt','.sql','xlsx','xls', ]
	for unwanted in unwanted_ends:
		if url.endswith(unwanted):
			return False
	return True 


from crawlfish import UrlsCrawler , ExcelSaveOption , CSVSaveOption ,JSONSaveOption ,JSONFileSaveOption
# The UrlsCrawler class lghtens the load for developers when scraping from a list of links 
# or from a text file containing the links 

# Lets say we have a function for scraping a url
def scrape_url(url):
	data = ["23" , "John" , "john@gmail.com"]
	return data



# now create the headers for our data 
headers = [ 'AGE' , 'NAME' , 'EMAIL' ]
urls = ["url1" , "url2","url3"] 



# create a save option to customize how you want your data saved
save_option = CSVSaveOption(file = "test_data/data.csv" , overwrite = True , delimiter = ",")
# lets get our data
data = UrlsCrawler().scrape_urls(  urls = urls  , scrape_function = scrape_url ,
		 data_headers = headers ,save_option =save_option)
# The data will be saved to a the file provided
with open("test_data/data.csv") as f:
	print(f.read())


#Scrape links from a file with the links line by line
links_file = "test_data/links.txt"
data  = UrlsCrawler().scrape_urls_in_file( links_file , scrape_function = scrape_url ,
			data_headers  =headers , save_option = save_option )


# Save to different formats
# you can use the different save options like so :

# excel spreasheet save option
ExcelSaveOption( "myfile.xlsx" , sheet_name = "sheetA" ,overwrite = False )

# to return the data as dict in the format { index_header:{key:data , key:data}........ }
# key_index is the index of an item in the scrape function result that will be used as the main key if the data is
# to be saved in a dictionary-like form
ave_option = JSONSaveOption(key_index = 2 )
data  = UrlsCrawler().scrape_urls_in_file( links_file , scrape_function = scrape_url ,
			data_headers  =headers , save_option = save_option )
data


# To save to  a .JSON file 
save_option =JSONFileSaveOption( 'test_data/myfile.json' , key_index = 2 , overwrite = True )
data  = UrlsCrawler().scrape_urls_in_file( links_file , scrape_function = scrape_url ,
			data_headers  =headers , save_option = save_option )
data


 

import setuptools
from setuptools import setup, find_packages

long_description = ""
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="crawlfish",
    version="0.1.4",
    author="Victor Kipkemboi",
    author_email="scriptilapia@gmail.com",
    long_description =long_description,
    long_description_content_type="text/markdown",
    description="""An easy to use web crawling library with handy functions for exploring whole websites or single webpages , dynamic element searching , generic web scraping  and  more . It also has helpful shortcut functions for bs4 , requests and selenium . """,
    url="https://github.com/victhepythonista/crawlfish",
    project_urls={
        "Bug Tracker": "https://github.com/victhepythonista/crawlfish/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License"
    ],
    install_requires = [
            'art',
            'ascii-magic',
            'openpyxl',
            'selenium' , 
            'requests', 
            'tabulate' ,
            'tldextract',
            'bs4',
            'tqdm',
            ],
    packages=find_packages(),
    python_requires=">=3.6",
 
)


import re , bs4
from bs4 import BeautifulSoup


def get_element_text(element:bs4.Tag , value_if_none:bool = ""):
	''' Get the text value of an element 

	Parameters
	----------
	element:bs4.Tag
		The element to extract text from 
	value_if_none:str
		What to return if text is not found 

	Returns
	str
		The text value found or value_if_none'''
	if element is None :
		return value_if_none
	else:
		st = element.string 
		if st:
			return st 
		tx = element.text 
		if tx:
			return tx 
	return value_if_none





import requests , socket , os  ,time 
import urllib3
from selenium import webdriver
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.remote.webdriver import WebDriver as BaseSeleniumWebDriver
from bs4 import BeautifulSoup
from crawlfish.crawlfish_utils import check_file

 
def get_url(url:str,   session:requests.Session = None ,
			 params:dict = {} ,
			  headers:dict = {}  , 
			  retry_timeout:int = 5 , 
			  max_retries:int = 2000 , 
			  retry_count:int = 0 
			   ):
	''' Fetches a url and returns the response object 

	Parameters
	----------
	url:str
		The url to fetch 
	session:requests.Session
		The session class to use to fetch the url 
	headers:dict
		The GET headers
	params:dict 
		The GET parameters 
	retry_timeout:int 
		Time in seconds to wait in between retries 
	max_retries:int 
		The maximum number of retries allowed
	retry_count:int
		The current retry count 


	Returns
	-------
	response:requests.Response 
		A response instance obtained after fetching the url 

	'''
	if not session:
		session = requests.Session()

	print(f"Getting page ->  {url} " , end = '')
	try:
		response = session.get(url , params = params , headers = headers)
		print("DONE")
		return response
	except (requests.exceptions.ConnectionError , 
		requests.exceptions.ConnectTimeout, 
		requests.exceptions.ChunkedEncodingError,
		socket.gaierror, requests.exceptions.ReadTimeout
		 ):
		print("Connect error . Retrying ....\n ")
		if retry_count > max_retries:
			print("-------------------------MAX RETRY COUNTS ACHIEVED -----------------------------")
			raise
		retry_count += 1
		print("         \n            [RETRY {}]  \n".format(retry_count))
		time.sleep(retry_timeout)
		return get_url(url , session , params , headers , retry_timeout , max_retries , retry_count)
	except:
		raise



def get_page(url , session:requests.Session = None ,
			 params = {} ,
			  headers = {}  , 
			  retry_timeout = 5 , 
			  max_retries = 2000 , 
			  retry_count = 0 , 
			  output_file = "", 
			  soup = False ):
	'''
	This function fetches a url and returns the html of the web page
	If there is a problem with the connection it will retry as specified

	Parameters
	----------
	url:str
		The url to fetch 
	session:requests.Session
		The session class to use to fetch the url 
	headers:dict
		The GET headers
	params:dict 
		The GET parameters 
	retry_timeout:int 
		Time in seconds to wait in between retries 
	max_retries:int 
		The maximum number of retries allowed
	retry_count:int
		The current retry count 
	soup:bool
		Whether or not to return the data a BeautifulSoup instance 
	output_file:str
		Where to save the HTML found



	Returns
	-------
	bs4.BeautifulSoup object or the html string of the page fetched 
		 
	
	'''
	response =get_url(url , session )
	soup  = BeautifulSoup(response.content , 'html.parser')
	cont = soup.prettify() 
	if output_file:
		print("Writing response content to file {}".format(output_file) , end = '')
		if not check_file(output_file):
			print("Could not write to file :( " , output_file)

		else:
			with open(output_file , "w" , encoding = 'utf-8') as f:
				f.write(cont)
			print('DONE')
		
	if soup:
		return soup
	return cont 

	 

def get_page_with_driver(
	url:str ,
	driver:BaseSeleniumWebDriver ,
	retry_timeout:int = 5 , 
	max_retries:int = 30 , 
	retry_count:int = 0 , 
	output_file:str = "" ):
	'''Try to get a web page using a selenium.webdriver.WebDriver object 

	Parameters
	----------
	driver:selenium.webdriver.WebDriver 
		The selenium driver to use for fetching the page
	url:str
		The url to fetch 
	retry_timeout:int 
		Time in seconds to wait in between retries 
	max_retries:int 
		The maximum number of retries allowed
	retry_count:int
		The current retry count 
	output_file:str
		Where to save the HTML found


	Returns
	-------
	html:str
		The html found after fetching the url ,an empty string if the page can't be reached
	'''
	print("Fetching page with driver " , driver , type(driver))
	html = ""
	if not isinstance(driver , BaseSeleniumWebDriver ):
		raise ValueError("Expected a webdriver.Remote object ")
	tolerable_exceptions = (urllib3.exceptions.ReadTimeoutError , TimeoutError , WebDriverException )
	try:
		driver.get(url)
	except tolerable_exceptions as e:
		print("FAILED")
		print("A network error occurred . Retrying to get url : " , url)
		print("Retrying after {} seconds ".format(retry_timeout))
		time.sleep(retry_timeout)
		if retry_count >= max_retries:
			print("========= TIMEOUT REACHED , Could not fetch url =========")
			return ''
		retry_count += 1
		return get_page_with_driver(url , driver , retry_timeout,
						max_retries , retry_count , output_file)
	except Exception as e:
		raise
	html = driver.page_source
	print("Page HTML fetched successfully")
	if output_file:
		print("Writing HTML to file {}".format(output_file) , end = '')
		if not check_file(output_file):
			print("Could not write HTML data to file ")
			return html 
		else:
			with open( output_file , "w" , encoding = "utf-8"  ) as f:
				f.write(html)
			print("DONE \n")
	return html





import os
from collections.abc import Iterable
import requests
import tldextract 
from tqdm import tqdm 
from crawlfish.crawlfish_utils import check_file ,check_folder
from .extract_file_info_from_url import extract_file_info_from_url




def download_video( 
	url:str ,
	output_folder:str = os.getcwd()   , 
	file:str = "" ,
	file_name:str = "",
	save_to_file:bool = True
	):
	'''This function downloads a video from the provided url  returns the bytes  data and save the data to a file if save_to_file=True
	Parameters
	----------
	url:str
		The URL of the video file to be  downloaded
	file:str
		The name of the file to write the downloaded data to 
	output_folder:str
		The folder/directory where to store the downloaded file 
	save_to_file:bool
		Whether or not to save the data to a file 
	file_name:
		The name of the file withholding the extension name where the downloaded data will be written .
		The extension will be added automatically if this is  provided 

	Returns
	-------
	tuple 
		(file_name,bytes_data .....)
	'''
	print("Downloading video {} --------- ".format(url) , end =''  )
	response = requests.get(url, stream = True)
	print("DONE") 
	content = response.content
	url_file_name ,extension = extract_file_info_from_url(url)
 
	if output_folder:
		output_folder = os.path.abspath(output_folder)
		check_folder(output_folder)
	if output_folder and not file and not file_name:
		file= os.path.join(output_folder , url_file_name )
	if file_name:
		file_name += extension
		file = os.path.join(output_folder , file_name  )
	if not file_name:
		file_name = os.path.split(file)[-1]
	if not save_to_file:
		return file_name , response.content 
	# write the data 
	if not check_file(file):
		return file_name , b''
	print("Saving file to " , file , '----------- ', end = '')
	with open(file, 'wb') as f: 
		for chunk in tqdm(response.iter_content(chunk_size = 1024*1024), 'Writing to file  {}'.format(file)): 
			if chunk: 
				f.write(chunk)
		print("DONE")
	return file_name , content


def download_videos(
	urls_list:Iterable ,
	output_folder:str = os.getcwd()
	    ):
	''' This functions downloads videos from a list of urls and return the data in a dict
	Parameters
	----------
	urls_list:Iterable,tuple,iter,set
		An iterable containing the urls of the text-based files to be downloaded
	output_folder:str
		The folder/directory where to store the downloaded files

	Returns
	-------
	files_data:dict
		{ filename:text_data , ...... }
		A dictionary containing the text data as values and file names as keys'''
	video_data = {}
	for url in urls_list:
		file_name , data = download_video(url , output_folder = output_folder)
		video_data[file_name] = data
	return video_data




import re , bs4
from collections.abc import Iterable 
from bs4 import BeautifulSoup
from .element_finder import ElementFinder
from .get_element_text import get_element_text

def get_elements_text (
	html_or_soup:tuple[str,BeautifulSoup] ,
	tag_name:str="" ,
	tag_names:Iterable=[] ,
	attribute:str = "" ,
	attributes:Iterable = [] ,
	attribute_value:str="" ,
	attribute_values:Iterable = [] , 
	text_match:str= "" ,
	text_matches:Iterable = []  
						  ):
	'''Returns the texts in a list of all the elements that match the provided specs 

	Parameters
	-----------
	html_or_soup:BeautifulSoup
		The html string or  soup to search from
	tag_name:str
		tag name to match
	tag_names:Iterable
		Tag names to match 
	attribute:str
		Element attribute name to match
	attributes:Iterable
		A list of attribute names to match 
	attribute_value:str
		The attribute value to match , it can also be a regex
	attribute_values:Iterable
		A list attribute values to match , they can also be a regex
	text_match:str
		The text value to match  , it can also be a regex 
	text_matches:Iterable
		A list of text valies to match against element text values . They can also be regex
	 
	Returns
	--------
	texts:list
		A list containing text data from the elements specified 
		Format [ text1 , text2 ..... ]
		'''
	elements = ElementFinder(html_or_soup).find_elements(tag_name, tag_names , attribute , 
												attributes , attribute_value , attribute_values , 
												text_match , text_matches )
	texts = [ get_element_text(element) for element in elements ]
	return texts
